@@ -28,8 +28,10 @@ use strict;
 use XAO::Utils;
 use XAO::Objects;
 
-use vars qw(@ISA);
-@ISA=XAO::Objects->load(objname => 'FS::Hash');
+use base XAO::Objects->load(objname => 'FS::Hash');
+
+use vars qw($VERSION);
+($VERSION)=(q$Id: Global.pm,v 1.2 2002/01/04 01:47:37 am Exp $ =~ /(\d+\.\d+)/);
 
 sub new ($%) {
     my $class=shift;
