@@ -46,10 +46,11 @@ package XAO::DO::FS::Glue;
 use strict;
 use XAO::Utils;
 use XAO::Objects;
-use XAO::Errors qw(XAO::DO::FS::Glue);
+
+use base XAO::Objects->load(objname => 'Atom');
 
 use vars qw($VERSION);
-($VERSION)=(q$Id: Glue.pm,v 1.7 2002/01/04 01:47:37 am Exp $ =~ /(\d+\.\d+)/);
+($VERSION)=(q$Id: Glue.pm,v 1.12 2002/03/19 03:37:08 am Exp $ =~ /(\d+\.\d+)/);
 
 ###############################################################################
 
@@ -138,12 +139,12 @@ sub new ($%) {
         $$self->{driver}=$driver;
 
         ##
-        # Checking if this is a request to trash everything and produce
+        # Checking if this is a request to trash everything and produce a
         # squeky clean new database.
         #
         if($args->{empty_database}) {
             $args->{empty_database} eq 'confirm' ||
-                $self->throw("new - request for 'empty_database' is not 'confirm'ed");
+                throw $self "new - request for 'empty_database' is not 'confirm'ed";
 
             $driver->initialize_database();
         }
@@ -160,7 +161,7 @@ sub new ($%) {
         # or from an object being cloned..
         #
         my $glue=ref($proto) ? $proto : $args->{glue};
-        $glue || $self->throw("new - required parameter missed 'glue'");
+        $glue || throw $self "new - required parameter missed 'glue'";
         $$self->{glue}=$glue;
     }
 
@@ -254,7 +255,9 @@ Rough equivalent of:
 
 sub destroy ($) {
     my $self=shift;
-    foreach my $key ($self->keys()) {
+    foreach my $key ($self->keys) {
+        my $type=$self->describe($key)->{type};
+        next if $type eq 'key';
         $self->delete($key);
     }
 }
@@ -274,7 +277,8 @@ database.
 
 sub disconnect () {
     my $self=shift;
-    $$self->{glue} && $self->throw("disconnect - only makes sense on database handler (did you mean 'detach'?)");
+    $$self->{glue} &&
+        throw $self "disconnect - only makes sense on database handler (did you mean 'detach'?)";
     if($$self->{driver}) {
         $$self->{driver}->disconnect();
         delete $$self->{driver};
@@ -303,7 +307,7 @@ sub fetch ($$) {
     # Normalizing and checking path
     #
     $path=$self->normalize_path($path);
-    substr($path,0,1) eq '/' || $self->throw("fetch - bad path ($path)");
+    substr($path,0,1) eq '/' || throw $self "fetch - bad path ($path)";
 
     ##
     # Going through the path and retrieving every element. Could be a
@@ -333,7 +337,7 @@ Returns relative object name that XAO::Objects would accept.
 
 sub objname ($) {
     my $self=shift;
-    $$self->{objname} || $self->throw("objname - must have an objname");
+    $$self->{objname} || throw $self "objname - must have an objname";
 }
 
 ###############################################################################
@@ -532,21 +536,6 @@ sub _driver ($) {
 
 ###############################################################################
 
-=item _empty_field ($)
-
-Cleans out given field
-
-=cut
-
-sub _empty_field ($$) {
-    my $self=shift;
-    my $name=shift;
-    my $table=$self->_class_description->{table};
-    $self->_driver->empty_field($table,$$self->{unique_id},$name);
-}
-
-###############################################################################
-
 =item _field_description ($)
 
 Returns the description of the given field.
@@ -561,6 +550,62 @@ sub _field_description ($$) {
 
 ###############################################################################
 
+=item _field_default ($;$) {
+
+Returns default value for the given field for a hash object. Also sets
+'default' in class description if it is not there -- this might happen
+when old database is used with new FS.
+
+Optional second argument is for optimization. It should hold field
+description hash reference if it is available in calling context.
+
+=cut
+
+sub _field_default ($$) {
+    my $self=shift;
+    my $field=shift;
+
+    my $desc=shift || $self->_field_description($field);
+
+#dprint "Checking default for '$field', stored=",$desc->{default};
+
+    return $desc->{default} if defined($desc->{default});
+
+    my $type=$desc->{type};
+    my $default;
+    if($type eq 'text' || $type eq 'words') {
+        $default='';
+    }
+    elsif($type eq 'integer' || $type eq 'real') {
+        if(!defined($desc->{minvalue}) && !defined($desc->{maxvalue})) {
+
+            # This happens only for 'real' type where we do not set any
+            # borders unless specified.
+            #
+            $default=0;
+        }
+        elsif($desc->{minvalue} <= 0 &&
+              (!defined($desc->{maxvalue}) || $desc->{maxvalue} >= 0)) {
+            $default=0;
+        }
+        else {
+            $default=$desc->{minvalue};
+        }
+    }
+    else {
+        eprint "Wrong type for trying to get default (name=$field, type=$type)";
+        $default='';
+    }
+
+    $desc->{default}=$default;
+
+#dprint "Checking default for '$field', calculated=",$desc->{default};
+
+    return $default;
+}
+
+###############################################################################
+
 =item _glue ()
 
 Returns glue object reference. Makes sense only in derived objects!
@@ -971,14 +1016,15 @@ sub _build_search_query ($%) {
                 $wrapped=1;
             }
 
-            my $base_name=$glue->upper_class($cn);
-            my $conn=$glue->_connector_name($cn,$base_name);
+            my $upper_class=$glue->upper_class($cn);
+            my $conn=$glue->_connector_name($cn,$upper_class);
             $conn=$classes{$cn} . '.' .
                   $self->_driver->mangle_field_name($conn);
+
             $clause.=" AND " if $clause;
-            $clause.="$previous->{index}.unique_id=$conn";
+            $clause.="$classes{$upper_class}.unique_id=$conn";
         }
-        $previous=$item;
+        $previous=1;
     }
 
     ##
@@ -1228,11 +1274,15 @@ sub _build_search_clause ($$$$$$) {
             $self->throw("Not implemented ws on words");
         }
         else {
-            $clause=$self->_driver->search_clause_ws($field,$rha,$rha_escaped);
+            my $tf;
+            ($clause,$tf)=$self->_driver->search_clause_ws($field,$rha,$rha_escaped);
             if(!$clause) {
                 $clause="$field LIKE '" . '%' . $rha_escaped . '%' . "'";
                 $$post_process=1;
             }
+            elsif(defined($tf)) {
+                push(@$values,$tf);
+            }
         }
     }
     elsif($op eq 'wq') {
@@ -1240,11 +1290,15 @@ sub _build_search_clause ($$$$$$) {
             $self->throw("Not implemented wq on words");
         }
         else {
-            $clause=$self->_driver->search_clause_wq($field,$rha,$rha_escaped);
+            my $tf;
+            ($clause,$tf)=$self->_driver->search_clause_wq($field,$rha,$rha_escaped);
             if(!$clause) {
                 $clause="$field LIKE '" . '%' . $rha_escaped . '%' . "'";
                 $$post_process=1;
             }
+            elsif(defined($tf)) {
+                push(@$values,$tf);
+            }
         }
     }
     else {
@@ -1373,7 +1427,7 @@ sub _add_data_placeholder ($%) {
     my $args=get_args(\@_);
 
     ##
-    # Temporary make words -> text.
+    # Temporarily make words -> text until words handling is fixed.
     #
     $args->{type}='text' if $args->{type} eq 'words';
 
@@ -1392,47 +1446,71 @@ sub _add_data_placeholder ($%) {
     undef $args;
 
     ##
-    # Whenever unique is set index is set too.
+    # Checking if this is a hash in a list stored in some other hash
+    # other then Global.
     #
-    $fdesc{index}=1 if $fdesc{unique};
+    my $connected=$self->upper_class eq 'FS::Global' ? 0 : 1;
 
+    ##
+    # Adding...
+    #
     if($type eq 'words') {
         $fdesc{maxlength}=100 unless $fdesc{maxlength};
-        $driver->add_field_text($table,$name,$fdesc{index},$fdesc{unique},$fdesc{maxlength});
+        $driver->add_field_text($table,$name,$fdesc{index},$fdesc{unique},
+                                $fdesc{maxlength},$connected);
         $driver->setup_dictionary($table,$name,$fdesc{maxlength});
     }
     elsif ($type eq 'text') {
         $fdesc{maxlength}=100 unless $fdesc{maxlength};
-        $driver->add_field_text($table,$name,$fdesc{index},$fdesc{unique},$fdesc{maxlength});
+        $driver->add_field_text($table,$name,$fdesc{index},$fdesc{unique},
+                                $fdesc{maxlength},$connected);
     }
     elsif ($type eq 'integer') {
         $fdesc{minvalue}=-0x80000000 unless defined($fdesc{minvalue});
         $fdesc{maxvalue}=0x7FFFFFFF unless defined($fdesc{maxvalue});
         $driver->add_field_integer($table,$name,$fdesc{index},$fdesc{unique},
-                                   $fdesc{minvalue},$fdesc{maxvalue});
+                                   $fdesc{minvalue},$fdesc{maxvalue},$connected);
     }
     elsif ($type eq 'real') {
         $fdesc{minvalue}+=0 if defined($fdesc{minvalue});
         $fdesc{maxvalue}+=0 if defined($fdesc{maxvalue});
         $driver->add_field_real($table,$name,$fdesc{index},$fdesc{unique},
-                                $fdesc{minvalue},$fdesc{maxvalue});
+                                $fdesc{minvalue},$fdesc{maxvalue},$connected);
     }
     else {
         $self->throw("_add_data_placeholder - unknown type ($type)");
     }
 
+    ##
+    # Updating in-memory data with our new field and setting default
+    # value for it.
+    #
+    $fdesc{default}=$self->_field_default($name,\%fdesc);
+    $desc->{fields}->{$name}=\%fdesc;
+
+    ##
+    # Checking default value length
+    #
+    if($type eq 'text' || $type eq 'words') {
+        my $dl=length($fdesc{default});
+        $dl <= 30 ||
+            throw $self "_add_data_placeholder - default text is longer then 30 characters";
+        $dl <= $fdesc{maxlength} ||
+            throw $self "_add_data_placeholder - default text is longer then maxlength ($fdesc{maxlength})";
+    }
+
     ##
     # Updating Global_Fields
     #
-    $desc->{fields}->{$name}=\%fdesc;
     $driver->store_row('Global_Fields',
                        'field_name',$name,
                        'table_name',$table,
                        { type => $fdesc{type},
-                         maxlength => $fdesc{maxlength},
                          index => $fdesc{unique} ? 2 : ($fdesc{index} ? 1 : 0),
-                         minvalue => $fdesc{minvalue},
+                         default => $fdesc{default},
+                         maxlength => $fdesc{maxlength},
                          maxvalue => $fdesc{maxvalue},
+                         minvalue => $fdesc{minvalue},
                        });
 }
 
@@ -1451,11 +1529,11 @@ sub _add_list_placeholder ($%) {
     my $name=$args->{name} || $self->throw("_add_list_placeholder - no 'name' argument");
     my $class=$args->{class} || $self->throw("_add_list_placeholder - no 'class' argument");
     my $key=$args->{key} || $self->throw("_add_list_placeholder - no 'key' argument");
-    $self->_check_name($key) || $self->throw("_add_list_placeholder - bad key name ($key)");
+    $self->check_name($key) || $self->throw("_add_list_placeholder - bad key name ($key)");
     my $connector;
     if($self->objname ne 'FS::Global') {  
         $connector=$args->{connector} || 'parent_unique_id';
-        $self->_check_name($connector) ||
+        $self->check_name($connector) ||
             $self->throw("_add_list_placeholder - bad connector name ($key)");
     }
 
@@ -1569,9 +1647,17 @@ sub _drop_data_placeholder ($$) {
     $uid || $self->throw("_drop_data_placeholder - no description for $table.$name in the Global_Fields");
     $driver->delete_row('Global_Fields',$uid);
 
-    delete $desc->{fields}->{$name};
+    ##
+    # If that field was marked as 'unique' or 'index' then we need to
+    # drop index as well.
+    #
+    my $connected=$self->upper_class eq 'FS::Global' ? 0 : 1;
+    my $index=$desc->{fields}->{$name}->{index};
+    my $unique=$desc->{fields}->{$name}->{unique};
 
-    $driver->drop_field($table,$name);
+    $driver->drop_field($table,$name,$index,$unique,$connected);
+
+    delete $desc->{fields}->{$name};
 }
 
 ##
@@ -1631,25 +1717,31 @@ sub _drop_list_placeholder ($$;$) {
 
 ###############################################################################
 
-=item _check_name ($)
+=item check_name ($)
 
 Checks if the given name is a valid field name to be used in put() or
 get(). Should not be overriden unless you fully understand potential
 effects.
 
-Valid name must start from letter and may consist from letters, digits
+Valid name must start from a letter and may consist from letters, digits
 and underscore symbol. Length is limited to 30 characters.
 
 Returns boolean value.
 
 =cut
 
-sub _check_name ($$) {
+sub check_name ($$) {
     my $self=shift;
     my $name=shift;
     defined($name) && $name =~ /^[a-z][a-z0-9_]*$/i && length($name)<=30;
 }
 
+sub _check_name ($$) {
+    my $self=shift;
+    dprint ref($self) . "_check_name - obsolete, please change to 'check_name'";
+    $self->check_name(@_);
+}
+
 sub normalize_path ($$) {
     my $self=shift;
     my $path=shift;
@@ -1658,11 +1750,6 @@ sub normalize_path ($$) {
     $path;
 }
 
-sub throw {
-    my $self=shift;
-    throw Error::Simple ref($self)."::".join('',@_,"\n");
-}
-
 ###############################################################################
 1;
 __END__
